############################################
# Purpose: This script can be used to take a namelist and adapt it once the code had the RPE implemented.
#
# Inputs:
#        -v , --vault  path to the vault generated when implementing the RPE.
#        namelists to modify, as arguments.
#        i.e. python AdaptNamelist.py -v vault.pkl namelist_cfg namelist_ref
#
#
############################################
from AutoRPE.UtilsRPE.SourceManager import load_vault
from AutoRPE.UtilsRPE.Getter import get_type_of_contents
import re

# Few functions needed
def line_is_assignation(_line):
    """
    Determine whether the string provided is a assignation.
    """
    _line = _line.strip()
    if not _line:
        return False
    first_character = _line[0]
    if first_character in ["!", "&"]:
        return False
    if _line.count("!"):
        _line = _line[:_line.find("!")]
    if _line.count("="):
        return True
    return False


def line_is_real_assignation(_line):
    """
    Determine whether the line is a real assignation.
    """
    if _line.count("!"):
        _line = _line[:_line.find("!")]
    value = _line.split("=")[1]
    if value.count(","):
        return False
    if value.count("'"):
        return False
    if not value.count("."):
        return False
    try:
        float(value)
        return True
    except ValueError:
        return False


if __name__ == "__main__":
    from optparse import OptionParser

    # Use OptionParser to parse command line arguments
    parser = OptionParser()
    parser.add_option("-v", "--vault", dest="vault", default=None, metavar="FILE",
                      help="Path to the vault database")

    (options, args) = parser.parse_args()
    vault_path = options.vault

    # Check that the vault_path is not None
    if vault_path is None:
        parser.print_help()
        exit(0)
    # Load vault
    vault = load_vault(vault_path)
    if vault is None:
        print("Not able to load %s !\n" % vault_path)
        exit(1)

    # Read paths to namelists from arguments provided
    if args == []:
        print("namelists to adapt should be provided as arguments")
        exit(1)
    all_namelists = args

    # Regex pattern used to catch the name of the variable which is being modified
    pattern = "^([^\!]*)=([^\!]*)"

    # Loop over all the namelists
    for namelist in all_namelists:
        # Open the file and read the content
        with open(namelist) as f:
            data = [line for line in f]
            # Loop over all the lines
            for index, _l in enumerate(data):
                # In case the line corresponds to an assignation
                if line_is_assignation(_l):
                    # Use regex to read the variable name and the assigned value
                    m = re.search(pattern, _l)
                    var, value = m.groups()
                    # Use the vault to determine the type of the variable
                    type_of_var = get_type_of_contents(var, procedure=None, vault=vault)

                    # In case the variable is a rpe_var, add a %val suffix
                    if type_of_var == "rpe_var":
                        var = "%s%%val" % var.strip()
                        data[index] = "%s = %s\n" % (var, value)
                    # In case the variable is a fld_n derived type, add a custom fix
                    if type_of_var == "fld_n":
                        split_value = value.split(",")
                        value = value.replace(split_value[1], "-1, " + split_value[1])
                        data[index] = "%s = %s" % (var, value)

            # Join all the lines
            text = "".join(data)
        # Overwrite the namelist with the fixes
        with open(namelist, "w") as f:
            f.write(text)
