import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
from AutoRPE.UtilsRPE.Inserter import change_code
from os.path import isfile, isdir


def get_command_line_arguments():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--variable--file-list", default=None,
                        help="Path to the filename with the list of variables.")
    parser.add_argument("-t", "--target-precision", default="dp",
                        help="Precision specification at which the variables in the list will be set")
    parser.add_argument("-w", "--working-precision", default="single",
                        help="Precision used for wp variables in the code. i.e. double for standard runs, "
                             "single for mixed-precision versions")
    parser.add_argument("-s", "--source", default=None,
                        help="Path to the source that will be modified.")
    parser.add_argument("-v", "--vault", default=None,
                        help="Path to the vault.")

    args = parser.parse_args()

    if args.variable_file_list is None:
        parser.error("Argument -l/--variable-file-list required!")
    if not isfile(args.variable_file_list):
        parser.error("Argument -l/--variable-list %s is not a file." % args.variable_file_list)
    if args.source is None:
        parser.error("Argument -s/--source required!")
    if not isdir(args.source):
        parser.error("Argument -s/--source  %s is not a valid directory." % args.source)
    if args.vault is None:
        parser.error("Argument -v/--vault required!")
    if not isfile(args.vault):
        parser.error("Argument -v/--vault %s is not a file." % args.vault)

    return args


if __name__ == "__main__":

    args = get_command_line_arguments()

    list_filename = args.variable_file_list
    vault_path = args.vault
    source_path = args.source
    target_precision = args.target_precision
    working_precision = args.working_precision

    # Setting the working precision
    VariablePrecision.wp = working_precision

    change_code(list_filename, source_path, vault_path, target_precision)
