import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.Cleaner as Cleaner
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern

import re
import warnings


def get_variable(contents, procedure=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions
    if not vault:
        vault = BasicFunctions.load_vault()

    contents = contents.replace(" ", "").strip()

    try:
        purged_contents = _clean_from_intrinsics(contents)
        contents = purged_contents
    except ValueError:
        print("Not able to clean content ", contents, " from intrinsics")

    contents = Cleaner.clean_contents(contents, vault)

    if re.search(r"^\w+$", contents):
        try:
            return vault.get_variable_by_name(var_name=contents, procedure=procedure)
        except Error.VariableNotFound:
            raise Error.VariableNotFound
    if contents.count("%val"):
        contents = contents.replace("%val", "")
        return get_variable(contents, procedure, vault)
    if contents.count("%"):
        try:
            return get_structure_member(contents, procedure, vault)
        except Error.MemberNotFound:
            raise Error.VariableNotFound

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        return get_variable(contents, procedure, vault)

    raise Error.VariableNotFound("Variable %s: not found" % contents)


def get_real_in_line(line, vault, current_block, check=False):
    # Scan the line to retrieve all real numbers used in routine calls or pointers assignation, and corresponding targets
    # Vectors to store return values
    all_var = []
    all_dummy_var = []

    # If check mode, returns procedures to fix and their arguments
    procedure_tf = []
    arguments_proc_tf = []
    argument_index_tf = []
    dummy_var_type_tf = []
    dummy_var_intent_tf = []

    # To store calls and called procedures names
    procedure_calls = []
    called_procedures = []

    # Search for calls to subroutines
    subroutine_calls = CallManager.call_to_subroutine(line)
    function_calls = CallManager.calls_to_functions(line, current_block.procedure, vault=vault)

    if check:
        null_return_val = [None, None, None, None, None]
    else:
        null_return_val = [None, None]

    if subroutine_calls:
        for sbr_call in subroutine_calls:
            called_subroutine = CallManager.find_called_subroutine(sbr_call)
            if called_subroutine == "date_and_time":
                continue
            procedure_calls.append(sbr_call)
            called_procedures.append(called_subroutine)
    elif function_calls:
        for function_call in function_calls:
            procedure_calls.append(function_call)
            called_procedures.append(CallManager.find_called_function(function_call))
    # Search for pointer assignment and returns if found
    elif PointerDealer.is_pointer_assignment(line):
        ptr, target = PointerDealer.get_pointer_target(line, current_block.procedure, vault)

        if [ptr, target] == [None, None] or ptr.type not in ["single", "double", "rpe_var"]:
            return null_return_val
        if ptr.type != target.type:
            message = "POINTER => " + ptr.name + " is " + ptr.type + " instead of  " + target.type + \
                      "in procedure" + current_block.procedure.name + "in file: " + current_block.module_name + "f90"
            if check:
                print(message)
            else:
                raise Error.PointerTarged(message)
        all_var.append(ptr)
        all_dummy_var.append(target)
    # Nothing found
    else:
        return null_return_val

    # If are there calls to procedures
    if called_procedures:
        for procedure_call, called_procedure in zip(procedure_calls, called_procedures):
            called_arguments = CallManager.find_call_arguments(procedure_call)
            if called_procedure in ["mpp_minloc", "mpp_maxloc"]:
                v, av = manage_loc_exception(called_arguments, vault, subprogram=current_block.procedure)
                all_var += v
                all_dummy_var += av
                continue
            if not called_arguments:
                continue
            try:
                procedure = vault.get_procedure_by_name(called_procedure, module_name=current_block.module_name)
            except Error.ProcedureNotFound:
                continue
            indices_to_ignore = []

            if isinstance(procedure, BasicStructures.Interface):
                argument_types = [get_type_of_contents(argument, procedure=current_block.procedure, vault=vault) for
                                  argument in
                                  called_arguments]
                argument_dimensions = [get_dimension_of_contents(argument, procedure=current_block.procedure, vault=vault)
                                       for argument
                                       in called_arguments]
                keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in
                                     called_arguments]
                # Checking if some of the arguments does not necessarily need to be bounded to the dummy
                # because there's already an existing version within the same interface with a different precision
                alternative_procedures = []
                for _in, _type in enumerate(argument_types):
                    new_argument_types = argument_types[:]
                    if _type == "single":
                        new_argument_types[_in] = "double"
                    elif _type == "double":
                        new_argument_types[_in] = "single"
                    else:
                        continue
                    try:
                        pp = get_interface_from_arguments_and_info(new_argument_types, procedure,
                                                                   argument_dimensions=argument_dimensions,
                                                                   keyword_arguments=keyword_arguments, vault=vault)
                        alternative_procedures.append(pp)
                        if not check:
                            indices_to_ignore.append(_in)
                    except AssertionError:
                        continue
                # Classical match
                try:
                    procedure = get_interface_from_arguments_and_info(argument_types, procedure,
                                                                      argument_dimensions=argument_dimensions,
                                                                      keyword_arguments=keyword_arguments, vault=vault)
                except AssertionError:
                    if check and not alternative_procedures:
                        # In case there wasn't a match with the interface, check if through a variable cast it can work
                        types_to_test = BasicFunctions.create_type_combinations(argument_types)
                        matches = []
                        for t in types_to_test:
                            try:
                                m = get_interface_from_arguments_and_info(
                                    t,
                                    procedure,
                                    argument_dimensions=argument_dimensions,
                                    keyword_arguments=keyword_arguments,
                                    vault=vault,
                                )
                                matches.append(m)
                                # break
                            except AssertionError:
                                continue
                        if len(matches) == 1:
                            procedure = matches[0]
                        elif len(matches) > 1:
                            print("Multi-match!")
                        else:
                            print("No match!")
                        if isinstance(procedure, BasicStructures.Interface):
                            print("INTERFACE: ", current_block.procedure.name, procedure.name,
                                  [at for at in argument_types], "file: ")
                            continue
                    elif check and alternative_procedures:
                        procedure = alternative_procedures[0]
                    else:
                        # TODO: This workaround here was required because the function dimension_of_contents still
                        # does not work properly.
                        # TODO: Probably not needed anymore
                        try:
                            procedure = get_interface_from_arguments_and_info(argument_types, procedure,
                                                                              argument_dimensions=None,
                                                                              keyword_arguments=keyword_arguments,
                                                                              vault=vault)
                        except AssertionError:
                            raise AssertionError
            if procedure is None:
                continue
            for index, argument in enumerate(called_arguments):
                if not check:
                    if index in indices_to_ignore:
                        continue
                key = None
                if is_keyword_argument(argument):
                    [key, argument] = argument.split("=", 1)
                    key = key.strip()
                    argument = argument.strip()
                    var_type = get_type_of_contents(argument, procedure=current_block.procedure, vault=vault)
                else:
                    var_type = get_type_of_contents(argument, procedure=current_block.procedure, vault=vault)

                if var_type not in ["single", "double", "rpe_var"] or NatureDeterminer.is_a_number(argument):
                    continue

                if key is not None:
                    dummy_var = procedure.get_variable_by_name(key)
                else:
                    dummy_var = procedure.get_variable_by_position(index)
                # In case of the argument being an input, avoid binding the variables.
                if not check and dummy_var.intent == "in":
                    continue
                if not check:
                    try:
                        var = get_variable(argument, procedure=current_block.procedure, vault=vault)
                        all_var.append(var)
                        all_dummy_var.append(dummy_var)
                    except AssertionError:
                        continue
                else:
                    if var_type != dummy_var.type:
                        var_name = Cleaner.clean_contents(argument)
                        print("ARGUMENT => ", var_name, " is ", var_type, " instead of  ", dummy_var.type,
                              "in call to", called_procedure, "file: ", current_block.procedure.module.name, ".f90")
                        procedure_tf.append(called_procedure)
                        arguments_proc_tf.append(argument)
                        argument_index_tf.append(index)
                        dummy_var_type_tf.append(dummy_var.type)
                        dummy_var_intent_tf.append(dummy_var.intent)

    if check:
        return [procedure_tf, arguments_proc_tf, argument_index_tf, dummy_var_type_tf, dummy_var_intent_tf]
    else:
        return [all_var, all_dummy_var]


# Retrieve variable corresponding to a structure member
def get_structure_member(contents, procedure, vault):
    root = contents.split("%")[0]
    try:
        root_variable = procedure.accessible_var[root][0]
    except KeyError:
        # Case of procedure or function with a use module, only var statement
        try:
            root_variable = [v for v in
                             procedure.use_only_variables
                             if v.name.lower() == root][0]
        except IndexError:
            raise Error.VariableNotFound
    try:
        root_derived_type = [dt for dt in vault.derived_types if dt.name.lower() == root_variable.type.lower()][0]
    except IndexError:
        try:
            root_derived_type = [dt for dt in vault.external_type if dt.lower() == root_variable.type.lower()][0]
            raise Error.ExternalTypeUsed("The external type %s has been used" % root_derived_type)
        except IndexError:
            raise Error.DerivedTypeNotFound
    layers = contents.split("%")[1:]
    for l in layers:
        try:
            layer = [v for v in root_derived_type.variables if v.name.lower() == l.lower()][0]
        except IndexError:
            # It means that mother type is an extended type
            try:
                layer = [v for v in root_derived_type.extension.variables if v.name.lower() == l.lower()][0]
            except IndexError:
                raise Error.MemberNotFound
            if layer:
                pass
            else:
                raise Error.VariableNotFound("Variable %s not found", contents)

        try:
            root_derived_type = [dt for dt in vault.derived_types if dt.name.lower() == layer.type.lower()][0]
        except IndexError:
            if l == contents.split("%")[-1]:
                pass
            else:
                raise Error.VariableNotFound("Variable %s not found", contents)

    return layer


def get_type_of_contents(contents, procedure=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions

    contents = contents.replace(" ", "").strip()
    contents_lower = contents.lower()

    # Basic check
    if contents[0] == "-" or contents[0] == "+":
        contents = contents[1:]

    if contents_lower[0] in ["'", '"'] or contents_lower[:4].strip() == 'trim':
        return "char"
    if contents.count("'") or contents.count("\""):
        return "char"
    if contents_lower.find("nf90_") == 0:
        return "integer"
    if contents_lower.find("set_nf90_") == 0:
        return "integer"
    is_a_comparison = Splitter.split_elements_in_comparison(contents)
    if is_a_comparison is not None:
        return "logical"
    if contents_lower.count(".false.") or contents_lower.count(".true."):
        return "logical"

    # Check if is an array
    try:
        # suppose that all the elements are of the same type
        contents = Splitter.split_hardcoded_array(contents)[0]
        return get_type_of_contents(contents, procedure, vault)
    except Error.IsNotAnArray:
        pass

    # After checking it is not an array we remove parenthesis
    if BasicFunctions.is_all_inside_parenthesis(contents):
        contents = contents.strip()[1:-1]

    # Case with call to a function with array arguments after call foo(a(1:3), b)
    match = re.search(r"(\w* *\([^=]*\)) *(\(.*\))", contents)
    if match:
        contents = match.group(1)

    # Check if is a call to function, and in this case if is external
    is_an_external_function = False
    if CallManager.is_call_to_function(contents, procedure, vault):
        try:
            return get_function_type(contents, procedure, vault)
        except Error.ProcedureNotFound:
            is_an_external_function = True

    if NatureDeterminer.has_operations(contents):
        split_elements = Splitter.split_elements_in_operations(contents, procedure=procedure, vault=vault)
        types = [get_type_of_contents(element, procedure, vault) for element in
                 split_elements]
        return BasicFunctions.combine_types(types)
    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        try:
            return get_type_of_contents(contents, procedure, vault)
        except Error.VariableNotFound:
            warnings.warn('External variable %s used' % contents)
            return None

    # Remove the index from array and DT:
    #   ght_abl <-- ght_abl(2:jpka)
    #   todelay%z1d <-- todelay(ji)%z1d(:)
    contents = Cleaner.clean_contents(contents, vault)
    if re.match("^" + RegexPattern.number_exp_DPE_notation + "$", contents):
        return VariablePrecision.wp

    if contents.count(","):
        # If there are commas at this point, it must correspond with a temporal array
        # Array elements should have the same type, so we will assume that all the elements
        # have the same type than the first element
        first_element = AttributeParser.attribute_split(contents)[0]
        return get_type_of_contents(first_element, procedure, vault)

    # Checks if it is a number
    num_type = NatureDeterminer.is_a_number(contents)
    if num_type:
        return num_type

    # Check if is a Derived data type
    if contents.count("%"):
        if contents.split("%")[-1] == "val":
            return VariablePrecision.dp
        try:
            member = get_structure_member(contents, procedure, vault)
        except Error.ExternalTypeUsed:
            return None
        return member.type

    # Search if is a simple variable
    if re.search(r"^\w+$", contents):
        try:
            var = vault.get_variable_by_name(var_name=contents, procedure=procedure)
            return var.type
        except Error.VariableNotFound:
            if contents_lower.count("mpi_"):
                return "integer"
            elif contents_lower == "rpe":
                return "rpe_var"
            elif is_an_external_function:
                warnings.warn("Not able to find type! :(  |%s|" % contents)
                return None
            else:
                try:
                    vault.get_procedure_by_name(contents)
                    # A function is being passed as an arg, as in CALL MPI_OP_CREATE(func,[...] )
                    return None
                except Error.ProcedureNotFound:
                    raise Error.VariableNotFound(contents)

    raise Error.TypeOfContent("Type of content for the argument %s: not found", contents)


def get_function_type(function_call, procedure, vault):
    fun_name = CallManager.find_called_function(function_call)
    fun_name = fun_name.lower()
    if fun_name in vault.intrinsics:
        if fun_name in ["reshape", "min", "max", "sum", "log", "log10", "maxval", "minval", "abs", "sqrt", "maxloc",
                        "minloc", "huge", "tiny", "transfer", "sign"]:
            return get_type_of_contents(CallManager.find_call_arguments(function_call)[0], procedure, vault)
        elif fun_name in ["int", "nint", "count", "size", "mod", "umask", "len", "len_trim", "count"]:
            return "integer"
        elif fun_name in ["real", "float", "sin", "cos", "acos", "exp", "cosh"]:
            if fun_name == "real":
                return get_type_of_real_cast(function_call)

            return VariablePrecision.wp
        elif fun_name in ["cmplx"]:
            return "complex"
        elif fun_name in ["trim", "ctime", "adjustl"]:
            return "char"
        elif fun_name in ["present"]:
            return "logical"
        elif fun_name in ["null", "c_loc"]:
            return "pointer"
        elif fun_name in ["rpe"]:
            return "rpe_var"
        else:
            raise ValueError("This intrinsic should be considered too: %s" % fun_name)
    else:
        try:
            fun = vault.get_procedure_by_name(fun_name)
            if isinstance(fun, BasicStructures.Interface):
                return fun.procedures[0].type
            else:
                return fun.type
        except Error.ProcedureNotFound:
            raise Error.ProcedureNotFound


def get_dimension_of_contents(contents, procedure=None, vault=None):
    # TODO: Improve this function to work properly in all cases.
    contents = BasicFunctions.remove_literals(contents)
    contents = contents
    # Its a Char
    if contents.count("'") or contents.count('"') or contents.lower().count("trim"):
        return 0
    if contents.count(".false.") or contents.count(".true."):
        return 0
    if is_keyword_argument(contents):
        return get_dimension_of_contents(contents.split("=")[1], procedure, vault)
    # Matches arrays in the form of (/a,b,c/) or [a,b,c]
    array_match = re.search(RegexPattern.hardcoded_array, contents, flags=re.I)
    if array_match:
        return 1
    # Adding a workaround for casts
    if CallManager.is_call_to_function(contents, procedure, vault):
        fun = CallManager.find_called_function(contents)
        if fun.lower() in ["real", "rpe"]:
            return get_dimension_of_contents(CallManager.find_call_arguments(contents)[0], procedure, vault)
        elif fun.lower() == "sum" or fun.lower() == "maxval":
            if contents.lower().count("dim="):
                dim = get_dimension_of_contents(CallManager.find_call_arguments(contents)[0], procedure, vault)
                if dim:
                    return dim - 1
                else:
                    return dim
            else:
                contents = CallManager.find_call_arguments(contents)[0]
        elif fun.lower() in ["sqrt", "min", "max"]:
            contents = CallManager.find_call_arguments(contents)[0]
        else:
            # it was not a cast?
            pass
    if NatureDeterminer.has_operations(contents):
        elements = Splitter.split_elements_in_operations(contents)
        dimensions = [get_dimension_of_contents(e, procedure=procedure, vault=vault) for e in elements]
        dimensions = [d for d in dimensions if d is not None]
        if not dimensions:
            dimensions = [0]
        return max(dimensions)
    if NatureDeterminer.is_a_number(contents):
        return 0
    if contents.count(":"):
        return contents.count(":")
        # Case with call to a function with array arguments after

    # Arrays with index specified
    if re.search(r"\w *\([0-9\w]+\)", contents):
        return 0
    try:
        var = get_variable(contents, procedure=procedure, vault=vault)
        # Checking if we have the var + slice specification, like sdjf%fdta(2,2,1,2)
        # then it must be an scalar
        if re.search(r"%s *\(.*\)" % var.name, contents):
            return 0
        return var.dimension
    except AssertionError:
        return None


def get_variable_id(vault, variable_name, procedure_name=None, module_name=None, ):
    # Small function created to easily find the id of variables that appear in the code
    vs = [v for v in vault.variables if v.name == variable_name]

    if module_name is not None:
        vs = [v for v in vs if v.procedure.module.name == module_name]
    if procedure_name is not None:
        vs = [v for v in vs if v.procedure.name == procedure_name]

    if len(vs) == 1:
        print("Variable found! ID: %i" % vs[0].id)
        return vs[0].id
    elif len(vs) > 1:
        print("More than one variable!")
        return None
    else:
        print("Not a single variable found!")
        return None


def get_interface_from_arguments(arguments, interface, procedure=None, vault=None):
    argument_types = [get_type_of_contents(argument, procedure, vault) for argument in arguments]
    argument_dim = [get_dimension_of_contents(argument, procedure, vault) for argument in arguments]
    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in arguments]
    return get_interface_from_arguments_and_info(argument_types, interface, argument_dim, keyword_arguments, vault)


def get_interface_from_arguments_and_info(argument_types, interface, argument_dimensions=None, keyword_arguments=None,
                                          vault=None):
    # This function is intended to take the argument info (types and dimensions) and return the subroutine
    # inside the interface that matches it.
    number_of_arguments = len(argument_types)
    if keyword_arguments is None:
        keyword_arguments = [False for _x in argument_types]
    for procedure in interface.subprogram:
        if not vault:
            dummy_arguments = procedure.get_dummy_arguments()
        else:
            dummy_arguments = [var for var in procedure.variables if var.is_dummy_argument and not var.is_member_of]

        try:
            mandatory_arguments = [argument for argument in dummy_arguments if not argument.is_optional]
        except AttributeError:
            print("A problem with dummy argument in procedure %s", procedure.name)
        if len(dummy_arguments) < number_of_arguments:
            continue
        if len(mandatory_arguments) > number_of_arguments:
            continue
        match = True
        for index, argument_type in enumerate(argument_types):
            if match is False:
                continue
            keyword = keyword_arguments[index]
            if keyword:
                var = procedure.get_variable_by_name(keyword)
                if var:
                    dummy_type = var.type
                    dummy_dimension = var.dimension
                else:
                    dummy_type = None
                    dummy_dimension = None
                    match = False
            else:
                dummy_type = procedure.get_variable_by_position(index).type
                dummy_dimension = procedure.get_variable_by_position(index).dimension
            if dummy_type != argument_type:
                if (dummy_type in VariablePrecision.real_types and argument_type == "rpe_var") or (
                        dummy_type == "rpe_var" and argument_type in VariablePrecision.real_types):
                    pass
                else:
                    match = False
            if argument_dimensions is not None:
                # At this point we'll just ignore char variables.
                if argument_type == "char":
                    continue
                argument_dimension = argument_dimensions[index]
                if argument_dimension is not None and argument_dimension != dummy_dimension:
                    match = False

        if match:
            return procedure
    raise AssertionError("No matching routine found for interface : %s" % interface.name)


def is_keyword_argument(argument):
    pattern = "^[^()'\"=]*=[^=]"
    if re.search(pattern, argument):
        return True
    else:
        return False


def get_type_of_real_cast(string):
    # Leave only the arguments
    arguments = CallManager.find_call_arguments(string)
    if len(arguments) == 1:
        return VariablePrecision.wp
    elif len(arguments) == 2:
        argument, type = arguments
        return VariablePrecision.lookup_table[type]


def _clean_from_intrinsics(contents, vault=None):
    # Try and remove intrinsics from around a variable
    if vault:
        intrinsics = vault.intrinsics
    else:
        intrinsics = SourceManager.list_of_intrinsics
    pattern = r"(\w+) *\("
    m = [_m for _m in re.finditer(pattern, contents)]
    ret_val = contents
    if m:
        for mat in reversed(m):
            function = contents[mat.start(1):mat.end(1)]
            if function in intrinsics:
                if function == "null":
                    continue
                arg_of_func = contents[mat.end(0):]

                try:
                    arg_of_intrinsic = BasicFunctions.close_brackets(arg_of_func)
                    # Bad workaround to fix the case in which the argument is not a vector
                    if arg_of_intrinsic.count(",("):
                        arg_of_intrinsic = arg_of_intrinsic.split(",")[0]
                    intrinsic_call = BasicFunctions.close_brackets(contents[mat.start(1):])
                except ValueError:
                    return contents
                ret_val = contents.replace(intrinsic_call, arg_of_intrinsic)
                if NatureDeterminer.has_intrinsics(ret_val):
                    try:
                        ret_val = _clean_from_intrinsics(ret_val)
                    except RecursionError:
                        return contents
                else:
                    return ret_val
            elif function == "real":
                arguments = CallManager.find_call_arguments(contents)
                return arguments[0]
    return ret_val


def manage_loc_exception(called_arguments, vault, subprogram):
    # When we have a mpp_XXXloc exception instead of linking arguments with dummies we'll link the arguments together
    called_vars = []
    for arg in called_arguments:
        try:
            var = get_variable(arg, procedure=subprogram, vault=vault)
            called_vars.append(var)
        except Error.VariableNotFound:
            continue
    argument_variables = []
    dummies = []
    for index1 in range(len(called_vars)):
        for index2 in range(index1 + 1, len(called_vars)):
            argument_variables.append(called_vars[index1])
            dummies.append(called_vars[index2])
    return argument_variables, dummies
